from collections import defaultdict
from contextlib import contextmanager
import io
import os
import stat
from unittest.mock import MagicMock

import pytest

from hop import models
from hop.io import StartPosition


# example GCN circular from https://gcn.gsfc.nasa.gov/gcn3_circulars.html
GCN_TITLE = "GCN GRB OBSERVATION REPORT"
GCN_NUMBER = "40"
GCN_SUBJECT = "GRB980329 VLA observations"
GCN_DATE = "98/04/03 07:10:15 GMT"
GCN_FROM = "Greg Taylor at NRAO"
GCN_BODY = """
G.B. Taylor, D.A. Frail (NRAO), S.R. Kulkarni (Caltech), and
the BeppoSAX GRB team report:

We have observed the field containing the proposed x-ray counterpart
1SAX J0702.6+3850 of GRB 980329 (IAUC 6854) with the VLA at 8.4 GHz
on UT 1998 Mar 30.2, April 1.1, and April 2.1.  Observations on April
1.1 detected a radio source VLA J0702+3850 within the 1 arcminute
error circle of 1SAX J0702.6+3850.  The coordinates of
VLA J0702+3850 are: ra = 07h02m38.02170s dec = 38d50'44.0170" (equinox
J2000) with an uncertainty of 0.05 arcsec in each coordinate.  The
size of this radio source is less than 0.25 arcsec.  The density of
sources on the sky stronger than 250 microJy at this frequency is
0.0145 arcmin**-2.

The flux density measurements of VLA J0702+3850 are as follows:

Date(UT)   8.4 GHz Flux Density
--------   ----------------------
Mar 30.2   166 +/- 50 microJy
Apr  1.1   248 +/- 16    "
Apr  2.1    65 +/- 25    "

where the uncertainty in the measurement reflects the 1 sigma rms
noise in the image.  These measurements clearly demonstrate that
the radio source is variable on timescales of less than 1 day.
This rapid variability is similar to that observed in the
radio afterglow from GRB 970508.  We propose VLA J0702+3850
is the radio afterglow from GRB 980329.

Additional radio observations are in progress.
"""

GCN_CIRCULAR = f"""\
TITLE:   {GCN_TITLE}
NUMBER:  {GCN_NUMBER}
SUBJECT: {GCN_SUBJECT}
DATE:    {GCN_DATE}
FROM:    {GCN_FROM}

{GCN_BODY}\
"""

VOEVENT_XML = """\
<?xml version='1.0' encoding='UTF-8'?>
<voe:VOEvent xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:voe="http://www.ivoa.net/xml/VOEvent/v2.0" xsi:schemaLocation="http://www.ivoa.net/xml/VOEvent/v2.0 http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd" version="2.0" role="observation" ivorn="ivo://gwnet/LVC#S200302c-1-Preliminary">
  <Who>
    <Date>2020-03-02T02:00:09</Date>
    <Author>
      <contactName>LIGO Scientific Collaboration and Virgo Collaboration</contactName>
    </Author>
  </Who>
  <What>
    <Param dataType="int" name="Packet_Type" value="150">
      <Description>The Notice Type number is assigned/used within GCN, eg type=150 is an LVC_PRELIMINARY notice</Description>
    </Param>
    <Param dataType="int" name="internal" value="0">
      <Description>Indicates whether this event should be distributed to LSC/Virgo members only</Description>
    </Param>
    <Param dataType="int" name="Pkt_Ser_Num" value="1">
      <Description>A number that increments by 1 each time a new revision is issued for this event</Description>
    </Param>
    <Param dataType="string" name="GraceID" ucd="meta.id" value="S200302c">
      <Description>Identifier in GraceDB</Description>
    </Param>
    <Param dataType="string" name="AlertType" ucd="meta.version" value="Preliminary">
      <Description>VOEvent alert type</Description>
    </Param>
    <Param dataType="int" name="HardwareInj" ucd="meta.number" value="0">
      <Description>Indicates that this event is a hardware injection if 1, no if 0</Description>
    </Param>
    <Param dataType="int" name="OpenAlert" ucd="meta.number" value="1">
      <Description>Indicates that this event is an open alert if 1, no if 0</Description>
    </Param>
    <Param dataType="string" name="EventPage" ucd="meta.ref.url" value="https://gracedb.ligo.org/superevents/S200302c/view/">
      <Description>Web page for evolving status of this GW candidate</Description>
    </Param>
    <Param dataType="string" name="Instruments" ucd="meta.code" value="H1,V1">
      <Description>List of instruments used in analysis to identify this event</Description>
    </Param>
    <Param dataType="float" name="FAR" ucd="arith.rate;stat.falsealarm" unit="Hz" value="9.349090689402942e-09">
      <Description>False alarm rate for GW candidates with this strength or greater</Description>
    </Param>
    <Param dataType="string" name="Group" ucd="meta.code" value="CBC">
      <Description>Data analysis working group</Description>
    </Param>
    <Param dataType="string" name="Pipeline" ucd="meta.code" value="gstlal">
      <Description>Low-latency data analysis pipeline</Description>
    </Param>
    <Param dataType="string" name="Search" ucd="meta.code" value="AllSky">
      <Description>Specific low-latency search</Description>
    </Param>
    <Group name="GW_SKYMAP" type="GW_SKYMAP">
      <Param dataType="string" name="skymap_fits" ucd="meta.ref.url" value="https://gracedb.ligo.org/api/superevents/S200302c/files/bayestar.fits.gz,0">
        <Description>Sky Map FITS</Description>
      </Param>
    </Group>
    <Group name="Classification" type="Classification">
      <Param dataType="float" name="BNS" ucd="stat.probability" value="0.0">
        <Description>Probability that the source is a binary neutron star merger (both objects lighter than 3 solar masses)</Description>
      </Param>
      <Param dataType="float" name="NSBH" ucd="stat.probability" value="0.0">
        <Description>Probability that the source is a neutron star-black hole merger (primary heavier than 5 solar masses, secondary lighter than 3 solar masses)</Description>
      </Param>
      <Param dataType="float" name="BBH" ucd="stat.probability" value="0.8895532192171397">
        <Description>Probability that the source is a binary black hole merger (both objects heavier than 5 solar masses)</Description>
      </Param>
      <Param dataType="float" name="MassGap" ucd="stat.probability" value="0.0">
        <Description>Probability that the source has at least one object between 3 and 5 solar masses</Description>
      </Param>
      <Param dataType="float" name="Terrestrial" ucd="stat.probability" value="0.11044678078286028">
        <Description>Probability that the source is terrestrial (i.e., a background noise fluctuation or a glitch)</Description>
      </Param>
      <Description>Source classification: binary neutron star (BNS), neutron star-black hole (NSBH), binary black hole (BBH), MassGap, or terrestrial (noise)</Description>
    </Group>
    <Group name="Properties" type="Properties">
      <Param dataType="float" name="HasNS" ucd="stat.probability" value="0.0">
        <Description>Probability that at least one object in the binary has a mass that is less than 3 solar masses</Description>
      </Param>
      <Param dataType="float" name="HasRemnant" ucd="stat.probability" value="0.0">
        <Description>Probability that a nonzero mass was ejected outside the central remnant object</Description>
      </Param>
      <Description>Qualitative properties of the source, conditioned on the assumption that the signal is an astrophysical compact binary merger</Description>
    </Group>
  </What>
  <WhereWhen>
    <ObsDataLocation>
      <ObservatoryLocation id="LIGO Virgo"/>
      <ObservationLocation>
        <AstroCoordSystem id="UTC-FK5-GEO"/>
        <AstroCoords coord_system_id="UTC-FK5-GEO">
          <Time unit="s">
            <TimeInstant>
              <ISOTime>2020-03-02T01:58:11.519119</ISOTime>
            </TimeInstant>
          </Time>
        </AstroCoords>
      </ObservationLocation>
    </ObsDataLocation>
  </WhereWhen>
  <Description>Report of a candidate gravitational wave event</Description>
  <How>
    <Description>Candidate gravitational wave event identified by low-latency analysis</Description>
    <Description>V1: Virgo 3 km gravitational wave detector</Description>
    <Description>H1: LIGO Hanford 4 km gravitational wave detector</Description>
  </How>
</voe:VOEvent>\
"""

MESSAGE_GCN_TEXT = \
    b'TITLE:            GCN/AMON NOTICE\n' \
    b'NOTICE_DATE:      Fri 12 Apr 24 05:34:35 UT\n' \
    b'NOTICE_TYPE:      ICECUBE Astrotrack Bronze \n' \
    b'STREAM:           25\n' \
    b'RUN_NUM:          139279\n' \
    b'EVENT_NUM:        10803235\n' \
    b'SRC_RA:           103.7861d {+06h 55m 09s} (J2000),\n' \
    b'                  104.1106d {+06h 56m 27s} (current),\n' \
    b'                  103.1176d {+06h 52m 28s} (1950)\n' \
    b'SRC_DEC:           +5.8716d {+05d 52\' 18"} (J2000),\n' \
    b'                   +5.8390d {+05d 50\' 20"} (current),\n' \
    b'                   +5.9364d {+05d 56\' 11"} (1950)\n' \
    b'SRC_ERROR:        69.22 [arcmin radius, stat-only, 90% containment]\n' \
    b'SRC_ERROR50:      26.96 [arcmin radius, stat-only, 50% containment]\n' \
    b'DISCOVERY_DATE:   20412 TJD;   103 DOY;   24/04/12 (yy/mm/dd)\n' \
    b'DISCOVERY_TIME:   20026 SOD {05:33:46.89} UT\n' \
    b'REVISION:         0\n' \
    b'ENERGY:           1.2126e+02 [TeV]\n' \
    b'SIGNALNESS:       3.0910e-01 [dn]\n' \
    b'FAR:              3.4096 [yr^-1]\n' \
    b'SUN_POSTN:         21.10d {+01h 24m 24s}   +8.87d {+08d 52\' 11"}\n' \
    b'SUN_DIST:          82.22 [deg]   Sun_angle= -5.5 [hr] (East of Sun)\n' \
    b'MOON_POSTN:        67.46d {+04h 29m 50s}  +26.13d {+26d 07\' 33"}\n' \
    b'MOON_DIST:         40.41 [deg]\n' \
    b'GAL_COORDS:       208.12,  3.50 [deg] galactic lon,lat of the event\n' \
    b'ECL_COORDS:       104.34,-16.88 [deg] ecliptic lon,lat of the event\n' \
    b'COMMENTS:         IceCube Bronze event.  \n' \
    b'COMMENTS:         The position error is statistical only, there is no systematic added.  '


MESSAGE_BLOB = "This is a sample blob message. It is unstructured and does not require special parsing."

MESSAGE_JSON = b'{"foo":"bar", "baz":5}'

MESSAGE_AVRO = \
    b'\x4f\x62\x6a\x01\x04\x14\x61\x76\x72\x6f\x2e\x63' \
    b'\x6f\x64\x65\x63\x08\x6e\x75\x6c\x6c\x16\x61\x76' \
    b'\x72\x6f\x2e\x73\x63\x68\x65\x6d\x61\x8c\x0b\x7b' \
    b'\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x72\x65\x63' \
    b'\x6f\x72\x64\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x61\x75\x74\x6f\x5f\x72\x65\x63\x6f' \
    b'\x72\x64\x32\x22\x2c\x20\x22\x66\x69\x65\x6c\x64' \
    b'\x73\x22\x3a\x20\x5b\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x22' \
    b'\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69\x74\x65' \
    b'\x6d\x73\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x7d' \
    b'\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x61' \
    b'\x72\x72\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x22\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69\x74' \
    b'\x65\x6d\x73\x22\x3a\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x22\x6d\x61\x70\x22\x2c\x20\x22\x76' \
    b'\x61\x6c\x75\x65\x73\x22\x3a\x20\x22\x6c\x6f\x6e' \
    b'\x67\x22\x7d\x7d\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x73\x75\x62\x5f\x6f\x62\x6a\x65\x63' \
    b'\x74\x73\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x22\x72\x65\x63\x6f\x72\x64\x22\x2c\x20\x22\x6e' \
    b'\x61\x6d\x65\x22\x3a\x20\x22\x61\x75\x74\x6f\x5f' \
    b'\x72\x65\x63\x6f\x72\x64\x31\x22\x2c\x20\x22\x66' \
    b'\x69\x65\x6c\x64\x73\x22\x3a\x20\x5b\x7b\x22\x74' \
    b'\x79\x70\x65\x22\x3a\x20\x22\x73\x74\x72\x69\x6e' \
    b'\x67\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20' \
    b'\x22\x66\x6f\x6f\x22\x7d\x2c\x20\x7b\x22\x74\x79' \
    b'\x70\x65\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x2c' \
    b'\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x62\x61' \
    b'\x72\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x22\x6e\x75\x6c\x6c\x22\x2c\x20\x22\x6e' \
    b'\x61\x6d\x65\x22\x3a\x20\x22\x62\x61\x7a\x22\x7d' \
    b'\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x7b' \
    b'\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x72\x65\x63' \
    b'\x6f\x72\x64\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x61\x75\x74\x6f\x5f\x72\x65\x63\x6f' \
    b'\x72\x64\x30\x22\x2c\x20\x22\x66\x69\x65\x6c\x64' \
    b'\x73\x22\x3a\x20\x5b\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x22\x62\x79\x74\x65\x73\x22\x2c\x20\x22' \
    b'\x6e\x61\x6d\x65\x22\x3a\x20\x22\x78\x65\x6e\x22' \
    b'\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x61\x72' \
    b'\x72\x61\x79\x22\x2c\x20\x22\x69\x74\x65\x6d\x73' \
    b'\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x7d\x2c\x20' \
    b'\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x68\x6f\x6d' \
    b'\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a' \
    b'\x20\x22\x64\x6f\x75\x62\x6c\x65\x22\x2c\x20\x22' \
    b'\x6e\x61\x6d\x65\x22\x3a\x20\x22\x64\x72\x65\x6c' \
    b'\x22\x7d\x5d\x7d\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x71\x75\x75\x78\x22\x7d\x5d\x7d\x2c' \
    b'\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x74\x68' \
    b'\x69\x6e\x67\x79\x22\x7d\x2c\x20\x7b\x22\x74\x79' \
    b'\x70\x65\x22\x3a\x20\x22\x62\x79\x74\x65\x73\x22' \
    b'\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x64' \
    b'\x61\x74\x61\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70' \
    b'\x65\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a' \
    b'\x20\x22\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69' \
    b'\x74\x65\x6d\x73\x22\x3a\x20\x22\x62\x6f\x6f\x6c' \
    b'\x65\x61\x6e\x22\x7d\x2c\x20\x22\x6e\x61\x6d\x65' \
    b'\x22\x3a\x20\x22\x6c\x6f\x67\x69\x63\x22\x7d\x5d' \
    b'\x7d\x00\xfd\x2b\x62\xfc\xdf\xb1\xd2\x03\x2e\x33' \
    b'\x42\xea\xa7\x2a\xc4\x52\x02\x70\x08\x02\x04\x06' \
    b'\x08\x00\x04\x04\x02\x61\x02\x02\x62\x04\x00\x04' \
    b'\x02\x63\x06\x02\x64\x08\x00\x00\x06\x61\x62\x63' \
    b'\x2c\x06\x64\x65\x66\x06\xb2\x01\x5c\x0a\x00\x58' \
    b'\x39\xb4\xc8\x76\xbe\x05\x40\x08\x41\x00\x42\x04' \
    b'\x04\x01\x00\x00\xfd\x2b\x62\xfc\xdf\xb1\xd2\x03' \
    b'\x2e\x33\x42\xea\xa7\x2a\xc4\x52'

# Equivalent to the data encoded in MESSAGE_AVRO
AVRO_DATA_EQUIVALENT = [{
    "arr": [1, 2, 3, 4],
    "sub_objects": [{"a": 1, "b": 2}, {"c": 3, "d": 4}],
    "thingy": {
        "foo": "abc",
        "bar": 22,
        "baz": None,
        "quux": {"xen": b"def", "hom": [89, 46, 5], "drel": 2.718}
    },
    "data": b"A\x00B\x04",
    "logic": [True, False],
}]

# This was the original configuration structure, which permitted only a single credential
AUTH_CONFIG_LEGACY = """\
[auth]
username = "username"
password = "password"
"""

# This is the new configuration structure, which contains a list of credentials
AUTH_CONFIG = """
auth = [{
         username="username",
         password="password"
         }]
"""

# This is the OIDC configuration structure, which is very similar
AUTH_CONFIG_OIDC = """
auth = [{
         username="username",
         password="password",
         token_endpoint="https://example.com/oauth2/token"
         }]
"""


class MockBroker:
    """Mock a Kafka broker.

    This stores internally messages and tracks offsets
    for different consumer groups.

    """

    def __init__(self):
        self.reset()

    def reset(self):
        self._messages = defaultdict(list)
        self._offsets = defaultdict(dict)

    def write(self, topic, msg, headers=[]):
        self._messages[topic].append((msg, headers))

    def has_message(self, topic, message, headers=[]):
        for m in self._messages[topic]:
            print(m[0] == message, m[1] == headers)
        return (message, headers) in self._messages[topic]

    def read(self, topics, groupid, start_at=StartPosition.EARLIEST, **kwargs):
        if isinstance(topics, str):
            topics = [topics]

        for topic in topics:
            if topic not in self._offsets or groupid not in self._offsets[topic]:
                if start_at == StartPosition.EARLIEST:
                    self._offsets[topic][groupid] = 0
                else:
                    self._offsets[topic][groupid] = len(self._messages[topic]) - 1

            try:
                offset = self._offsets[topic][groupid]
                self._offsets[topic][groupid] += 1
                yield from self._messages[topic][offset:]
            except IndexError:
                pass


@pytest.fixture(scope="session")
def mock_broker():
    return MockBroker()


@pytest.fixture(scope="session")
def mock_producer():
    def _mock_producer(mock_broker, topic):
        class ProducerBrokerWrapper:
            def __init__(self, broker, topic):
                self.broker = broker
                self.topic = topic

            def write(self, msg, headers=[], delivery_callback=None, topic=None):
                if topic is None:
                    if self.topic is not None:
                        topic = self.topic
                    else:
                        raise Exception("No topic specified for write")
                self.broker.write(topic, msg, headers)

            def close(self):
                pass

            def __enter__(self):
                return self

            def __exit__(self, *exc):
                pass

        producer = ProducerBrokerWrapper(mock_broker, topic)
        return producer

    return _mock_producer


@pytest.fixture(scope="session")
def mock_consumer():
    def _mock_consumer(mock_broker, topics, group_id, start_at=StartPosition.EARLIEST):
        class ConsumerBrokerWrapper:
            def __init__(self, broker, topics, group_id, start_at):
                if isinstance(topics, str):
                    topics = [topics]
                self.broker = broker
                self.topics = topics
                self.group_id = group_id
                self.start_at = start_at

            def subscribe(self, topics):
                for topic in topics:
                    assert topic in set(self.topics)

            def stream(self, *args, **kwargs):
                class Message:
                    def __init__(self, value, headers=[]):
                        self._value = value
                        self._headers = headers

                    def value(self):
                        return self._value

                    def headers(self):
                        return self._headers

                for message in self.broker.read(self.topics, self.group_id, self.start_at, **kwargs):
                    yield Message(message[0], message[1])

            def close(self):
                pass

            def __enter__(self):
                return self

            def __exit__(self, *exc):
                pass

        consumer = ConsumerBrokerWrapper(mock_broker, topics, group_id, start_at)
        return consumer

    return _mock_consumer


@pytest.fixture(scope="session")
def mock_kafka_message():
    message = MagicMock()
    message.topic.return_value = "test-topic"
    message.partition.return_value = 0
    message.offset.return_value = 0
    message.timestamp.return_value = (0, 1234567890)
    message.key.return_value = "test-key"
    message.headers.return_value = [("a header", "a value"), ("another header", "other value")]
    return message


@pytest.fixture(scope="session")
def legacy_auth_config():
    return AUTH_CONFIG_LEGACY


@pytest.fixture(scope="session")
def auth_config():
    return AUTH_CONFIG


@pytest.fixture(scope="session")
def auth_config_oidc():
    return AUTH_CONFIG_OIDC


@pytest.fixture(scope="session")
def circular_data_raw():
    return {
        "header": {
            "title": GCN_TITLE,
            "number": GCN_NUMBER,
            "subject": GCN_SUBJECT,
            "date": GCN_DATE,
            "from": GCN_FROM,
        },
        "body": GCN_BODY,
    }


@pytest.fixture(scope="session")
def circular_text():
    return GCN_CIRCULAR


@pytest.fixture(scope="session")
def circular_msg():
    return models.GCNCircular.load(GCN_CIRCULAR)


@pytest.fixture(scope="session")
def avro_data_raw():
    return AVRO_DATA_EQUIVALENT


@pytest.fixture(scope="session")
def avro_data():
    return MESSAGE_AVRO


@pytest.fixture(scope="session")
def avro_msg():
    return models.AvroBlob.load(MESSAGE_AVRO)


@pytest.fixture(scope="session")
def voevent_fileobj():
    return io.BytesIO(VOEVENT_XML.encode())


@pytest.fixture(scope="session")
def voevent_text():
    return VOEVENT_XML


@pytest.fixture(scope="session")
def gcn_text_notice_fileobj():
    return io.BytesIO(MESSAGE_GCN_TEXT)


@pytest.fixture(scope="session")
def gcn_text_notice_data():
    return MESSAGE_GCN_TEXT


@pytest.fixture(scope="session")
def blob_text():
    return MESSAGE_BLOB


@pytest.fixture(scope="session")
def blob_msg():
    return {"content": MESSAGE_BLOB}


message_parameters_dict_data = {
    # Generalize model_name, expected_model, test_file, and model_text
    # for easy access during tests. Useful when combined with parametrization
    # across message format, since fixtures (e.g., GCN_CIRCULAR) cannot be
    # used as parametrize arguments.
    "circular": {
        "model_name": "GCNCircular",
        "expected_model": models.GCNCircular,
        "test_file": "example_gcn.gcn3",
        "model_text": GCN_CIRCULAR,
    },
    "voevent": {
        "model_name": "VOEvent",
        "expected_model": models.VOEvent,
        "test_file": "example_voevent.xml",
        "model_text": VOEVENT_XML.encode(),
    },
    "gcn_text_notice": {
        "model_name": "GCNTextNotice",
        "expected_model": models.GCNTextNotice,
        "test_file": "example_gcn.txt",
        "model_text": MESSAGE_GCN_TEXT,
    },
    "json": {
        "model_name": "JSONBlob",
        "expected_model": models.JSONBlob,
        "test_file": "example_json.json",
        "model_text": MESSAGE_JSON,
    },
    "blob": {
        "model_name": "Blob",
        "expected_model": models.Blob,
        "test_file": "example_blob.txt",
        "model_text": MESSAGE_BLOB,
    },
    "avro": {
        "model_name": "AvroBlob",
        "expected_model": models.AvroBlob,
        "test_file": "example_avro.avro",
        "model_text": MESSAGE_AVRO,
    },
}


@pytest.fixture(scope="session")
def message_parameters_dict():
    return message_parameters_dict_data


@contextmanager
def temp_environ(**vars):
    """
    A simple context manager for temporarily setting environment variables

    Kwargs:
        variables to be set and their values

    Returns:
        None
    """
    from os import environ
    original = dict(environ)
    os.environ.update(vars)
    try:
        yield  # no value needed
    finally:
        # restore original data
        os.environ.clear()
        os.environ.update(original)


@contextmanager
def temp_config(tmpdir, data, perms=stat.S_IRUSR | stat.S_IWUSR):
    """
    A context manager which creates a temporary config file with specified data and permissions

    Args:
        data: the data to be written to the file
        perms: the permissions which should be set on the file.
            The default value is to use the standard, safe permissions

    Returns:
        The path to the config directory for hop to use this config file, as a string
    """

    config_path = f"{tmpdir}/hop/auth.toml"
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    config_file = open(config_path, mode='w')
    os.chmod(config_path, perms)
    config_file.write(data)
    config_file.close()
    try:
        yield str(tmpdir)
    finally:
        # remove file
        os.remove(config_path)



import logging
import os

from . import cli


logger = logging.getLogger("hop")


def get_config_path(type: str = "general"):
    """Determines the default location for auth configuration.

    Args:
        type: The type of configuration data for which the path should be looked up.
              Recognized types are 'general' and 'auth'.

    Returns:
        The path to the requested configuration file.

    Raises:
        ValueError: Unrecognized config type requested.

    """

    file_for_type = {
        "general": "config.toml",
        "auth": "auth.toml",
    }

    if type not in file_for_type:
        raise ValueError(f"Unknown configuration type: '{type}'")

    auth_filepath = os.path.join("hop", file_for_type[type])
    if "XDG_CONFIG_HOME" in os.environ:
        return os.path.join(os.getenv("XDG_CONFIG_HOME"), auth_filepath)
    else:
        return os.path.join(os.getenv("HOME"), ".config", auth_filepath)


def _add_parser_args(parser):
    cli.add_logging_opts(parser)

    subparser = parser.add_subparsers(title="commands", metavar="<command>", dest="command")
    subparser.required = True

    locate_parser = subparser.add_parser("locate", help="display configuration path")
    locate_parser.add_argument("-t", "--type", dest="type", default="general", required=False,
                               choices=["general", "auth"],
                               help="The type of configuration file to locate. "
                               "Defaults to general configuration.")


def _main(args):
    """Configuration utilities.

    """
    cli.set_up_logger(args)

    if args.command == "locate":
        print(get_config_path(args.type))



import pytest
import os
from conftest import temp_environ

from hop import configure


def test_get_config_path(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        if "XDG_CONFIG_HOME" in os.environ:
            # this change will revert at the end of the with block
            del os.environ["XDG_CONFIG_HOME"]

        # with HOME set but not XDG_CONFIG_HOME the config location should resolve to inside
        # ${HOME}/.config
        expected_path = os.path.join(tmpdir, ".config", "hop", "config.toml")
        config_loc = configure.get_config_path()
        assert config_loc == expected_path

        with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
            # with XDG_CONFIG_HOME set, no .config path component should be assumed
            expected_path = os.path.join(tmpdir, "hop", "config.toml")
            config_loc = configure.get_config_path()
            assert config_loc == expected_path


def test_get_config_path_auth(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        if "XDG_CONFIG_HOME" in os.environ:
            # this change will revert at the end of the with block
            del os.environ["XDG_CONFIG_HOME"]

        # with HOME set but not XDG_CONFIG_HOME the config location should resolve to inside
        # ${HOME}/.config
        expected_path = os.path.join(tmpdir, ".config", "hop", "auth.toml")
        config_loc = configure.get_config_path("auth")
        assert config_loc == expected_path

        with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
            # with XDG_CONFIG_HOME set, no .config path component should be assumed
            expected_path = os.path.join(tmpdir, "hop", "auth.toml")
            config_loc = configure.get_config_path("auth")
            assert config_loc == expected_path


def test_get_config_path_invalid(tmpdir):
    with pytest.raises(ValueError):
        configure.get_config_path("hairstyle")


def test_no_command_configure(script_runner):
    warning_message = (
        "hop configure: error: the following arguments are required: <command>"
    )
    ret = script_runner.run("hop", "configure")
    assert not ret.success
    assert warning_message in ret.stderr



from collections import Counter
from dataclasses import asdict, fields
import json
import logging
from pathlib import Path
import time
from unittest.mock import patch, MagicMock
from uuid import uuid4

import pytest

from hop.auth import Auth, AmbiguousCredentialError
from hop import io
from hop.models import AvroBlob, Blob, GCNCircular, GCNTextNotice, JSONBlob, VOEvent, format_name
import confluent_kafka

from conftest import temp_environ, temp_config, message_parameters_dict_data

logger = logging.getLogger("hop")


def content_mock(message_model):
    """Mock content to pass during the message_model creation since MagicMock()
    is unable to mock __init__ of the model dataclass in tests.
    """
    content = {}
    for field in fields(message_model):
        content.update({field.name: "test"})
    return content


def make_message(content, headers=[], topic="test-topic", partition=0, offset=0):
    message = MagicMock()
    message.value.return_value = content
    message.headers.return_value = headers
    message.topic.return_value = topic
    message.partition.return_value = partition
    message.offset.return_value = offset
    message.timestamp.return_value = (0, 1234567890)
    message.key.return_value = "test-key"
    return message


def make_message_standard(message, **kwags):
    raw = message.serialize()
    return make_message(raw["content"],
                        headers=[("_format", raw["format"].encode("utf-8"))], **kwags)


# only applicable to old message models which are JSON-compatible
def old_style_message(message):
    raw = {"format": format_name(type(message)), "content": asdict(message)}
    return json.dumps(raw).encode("utf-8")


def get_model_data(model_name):
    return message_parameters_dict_data[model_name]["model_text"]


# test the deserializer for each message format
@pytest.mark.parametrize("message", [
    # properly formatted, new-style messages
    {"format": "voevent",
        "content": make_message_standard(VOEvent.load(get_model_data("voevent")))},
    {"format": "gcn_text_notice",
        "content": make_message_standard(GCNTextNotice.load(get_model_data("gcn_text_notice")))},
    {"format": "circular",
        "content": make_message_standard(GCNCircular.load(get_model_data("circular")))},
    {"format": "json", "content": make_message_standard(JSONBlob.load(get_model_data("json")))},
    {"format": "avro", "content": make_message_standard(AvroBlob.load(get_model_data("avro")))},
    {"format": "blob", "content": make_message_standard(Blob(b"some data"))},
    # a new-style message in some user-defined format we don't have loaded
    {"format": "other", "content": make_message(b"other", headers=[("_format", b"other")])},
    # valid, old-style messages
    {"format": "voevent",
        "content": make_message(old_style_message(VOEvent.load(get_model_data("voevent"))))},
    {"format": "circular",
        "content": make_message(old_style_message(GCNCircular.load(get_model_data("circular"))))},
    # an old-style message with the old JSON label
    {"format": "json",
        "content": make_message(b'{"format":"blob", "content":{"foo":"bar", "baz":5}}')},
    # messages produced by foreign clients which don't apply our format labels
    {"format": "json", "content": make_message(b'{"foo":"bar", "baz":5}')},
    {"format": "blob",
        "content": make_message(b'some arbitrary data\xDE\xAD\xBE\xEFthat hop can\'t read')},
])
def test_deserialize(message, message_parameters_dict, caplog):

    message_format = message["format"]
    message_content = message["content"]

    # load parameters from conftest for valid formats
    if message_format in message_parameters_dict:
        message_parameters = message_parameters_dict[message_format]
        model_name = message_parameters["model_name"]
        expected_model = message_parameters["expected_model"]

        # test valid formats
        test_model = io.Deserializer.deserialize(message_content)

        # verify the message is classified properly
        assert isinstance(test_model, expected_model)

    else:  # test an invalid format
        with caplog.at_level(logging.WARNING):
            test_model = io.Deserializer.deserialize(message_content)

            # verify a message blob was produced with warnings
            output = f"Message format {message_format.upper()} " \
                "not recognized; returning a Blob"
            assert isinstance(test_model, Blob)
            assert output in caplog.text


def test_stream_read(circular_msg):
    topic = "gcn"
    group_id = "test-group"
    start_at = io.StartPosition.EARLIEST

    fake_message = make_message(
        GCNCircular.load(get_model_data("circular")).serialize()["content"],
        headers=[("_format", b"circular")])

    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[fake_message])
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url1 = f"kafka://hostname:port/{topic}"
        broker_url2 = f"kafka://{group_id}@hostname:port/{topic}"

        messages = 0
        with stream.open(broker_url1, "r") as s:
            for msg in s:
                assert isinstance(msg, GCNCircular), \
                    "Message should be deserialized to the expected model type"
                messages += 1
        assert messages == 1

        messages = 0
        with stream.open(broker_url2, "r") as s:
            for msg in s:
                messages += 1
        assert messages == 1


def test_stream_read_test_channel(circular_msg):
    start_at = io.StartPosition.EARLIEST
    message_data = {"format": "circular", "content": circular_msg}
    fake_message = make_message_standard(circular_msg)

    def test_headers():
        return [('_test', b'true')]

    fake_message.headers = test_headers
    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[fake_message])
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url = "kafka://hostname:port/test-topic"
        messages = 0
        with stream.open(broker_url, "r") as s:
            for msg in s:
                messages += 1
        assert messages == 0

        messages = 0
        with stream.open(broker_url, "r", ignoretest=False) as s:
            for msg in s:
                messages += 1
        assert messages == 1


def test_stream_read_multiple(circular_msg):
    group_id = "test-group"
    topic1 = "gcn1"
    topic2 = "gcn2"
    start_at = io.StartPosition.EARLIEST

    message_data = {"format": "circular", "content": circular_msg}
    topic1_message = make_message_standard(circular_msg, topic=topic1)
    topic2_message = make_message_standard(circular_msg, topic=topic2)
    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[topic1_message, topic2_message])
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url = f"kafka://{group_id}@hostname:port/{topic1},{topic2}"

        messages = Counter()
        with stream.open(broker_url, "r") as s:
            for msg, meta in s.read(metadata=True):
                messages[meta.topic] += 1

        assert messages[topic1] == 1
        assert messages[topic2] == 1


def test_stream_read_raw(circular_msg):
    topic = "gcn"
    group_id = "test-group"
    start_at = io.StartPosition.EARLIEST

    fake_message = make_message(
        GCNCircular.load(get_model_data("circular")).serialize()["content"],
        headers=[("_format", b"circular")])

    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[fake_message])
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url1 = f"kafka://hostname:port/{topic}"
        broker_url2 = f"kafka://{group_id}@hostname:port/{topic}"

        messages = 0
        with stream.open(broker_url1, "r") as s:
            for msg in s.read_raw():
                print(msg)
                # Since the message should be raw, we should be able to
                # deserialize it using the correct model class.
                des = GCNCircular.deserialize(msg)
                assert des
                messages += 1
        assert messages == 1

        messages = 0
        with stream.open(broker_url2, "r") as s:
            for result in s.read_raw(metadata=True):
                assert len(result) == 2, "Result should be a 2-tuple"
                msg = result[0]
                metadata = result[1]
                des = GCNCircular.deserialize(msg)
                assert des
                messages += 1
        assert messages == 1


def test_stream_read_test_raw(circular_msg):
    start_at = io.StartPosition.EARLIEST
    message_data = {"format": "circular", "content": circular_msg}
    fake_message = make_message_standard(circular_msg)

    def test_headers():
        return [('_test', b'true')]

    fake_message.headers = test_headers
    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[fake_message])
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url = "kafka://hostname:port/test-topic"
        messages = 0
        with stream.open(broker_url, "r") as s:
            for msg in s.read_raw():
                messages += 1
        assert messages == 0

        messages = 0
        with stream.open(broker_url, "r", ignoretest=False) as s:
            for msg in s.read_raw():
                messages += 1
        assert messages == 1


def test_stream_stop(circular_msg):
    start_at = io.StartPosition.EARLIEST
    fake_message = make_message_standard(circular_msg)

    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[fake_message])
    mock_instance.stop = MagicMock()
    stream = io.Stream(start_at=start_at, until_eos=True, auth=False)
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        broker_url = "kafka://hostname:port/gcn"

        with stream.open(broker_url, "r") as s:
            for msg in s:
                assert isinstance(msg, GCNCircular), \
                    "Message should be deserialized to the expected model type"
            s.stop()
        mock_instance.stop.assert_called_once()


def test_stream_write(circular_msg, circular_text, mock_broker, mock_producer):
    topic = "gcn"
    expected_msg = make_message_standard(circular_msg)
    fixed_uuid = uuid4()

    auth = Auth("user", "password")
    headers = {"some header": b"some value", "another header": b"other value"}
    canonical_headers = [("some header", b"some value"),
                         ("another header", b"other value"),
                         ("_id", fixed_uuid.bytes),
                         ("_sender", auth.username.encode("utf-8")),
                         ("_format", b"circular")]
    test_headers = canonical_headers.copy()
    test_headers.insert(4, ('_test', b'true'))
    none_test_headers = [("_id", fixed_uuid.bytes), ("_sender", auth.username.encode("utf-8")),
                         ('_test', b"true"), ("_format", b"circular")]

    mb = mock_broker
    with patch("hop.io.producer.Producer", side_effect=lambda c: mock_producer(mb, c.topic)), \
            patch("hop.io.uuid4", MagicMock(return_value=fixed_uuid)):

        broker_url = f"kafka://localhost:port/{topic}"
        start_at = io.StartPosition.EARLIEST
        until_eos = True

        stream = io.Stream(start_at=start_at, until_eos=until_eos, auth=auth)

        # verify warning is raised when groupid is set in write mode
        with pytest.warns(UserWarning):
            stream.open("kafka://localhost:9092/topic1", "w", group_id="group")

        mock_broker.reset()
        with stream.open(broker_url, "w") as s:
            s.write(circular_msg, headers)
            assert mock_broker.has_message(topic, expected_msg.value(), canonical_headers)

        mock_broker.reset()
        with stream.open(broker_url, "w") as s:
            s.write(circular_msg, headers, test=True)
            assert mock_broker.has_message(topic, expected_msg.value(), test_headers)

        mock_broker.reset()
        with stream.open(broker_url, "w") as s:
            s.write(circular_msg, headers=None, test=True)
            assert mock_broker.has_message(topic, expected_msg.value(), none_test_headers)

        # repeat, but with a manual close instead of context management
        mock_broker.reset()
        s = stream.open(broker_url, "w")
        s.write(circular_msg, headers)
        s.close()
        assert mock_broker.has_message(topic, expected_msg.value(), canonical_headers)

        mock_broker.reset()
        # more than one topics should now be allowed in write mode
        with stream.open("kafka://localhost:9092/topic1,topic2", "w") as s:
            with pytest.raises(Exception):
                # however, a topic must be specified when calling write with multiple topics
                # specified on construction
                s.write(circular_msg, headers)

            # selecting a topic explicitly when calling write should work
            s.write(circular_msg, headers, topic="topic1")
            assert mock_broker.has_message("topic1", expected_msg.value(), canonical_headers)
            s.write(circular_msg, headers, topic="topic2")
            assert mock_broker.has_message("topic2", expected_msg.value(), canonical_headers)

        mock_broker.reset()
        # no topic can also be specified in write mode
        with stream.open("kafka://localhost:9092/", "w") as s:
            with pytest.raises(Exception):
                # however, a topic must be specified when calling write
                s.write(circular_msg, headers)

            s.write(circular_msg, headers, topic="topic1")
            assert mock_broker.has_message("topic1", expected_msg.value(), canonical_headers)
            s.write(circular_msg, headers, topic="topic2")
            assert mock_broker.has_message("topic2", expected_msg.value(), canonical_headers)


def test_stream_write_raw(circular_msg, circular_text, mock_broker, mock_producer):
    topic = "gcn"
    encoded_msg = io.Producer.pack(circular_msg)
    headers = {"some header": "some value"}
    canonical_headers = list(headers.items())
    mb = mock_broker
    with patch("hop.io.producer.Producer", side_effect=lambda c: mock_producer(mb, c.topic)):
        stream = io.Stream(auth=False)

        broker_url = f"kafka://localhost:9092/{topic}"

        mock_broker.reset()
        with stream.open(broker_url, "w") as s:
            s.write_raw(encoded_msg, canonical_headers)
            assert mock_broker.has_message(topic, encoded_msg, canonical_headers)

        # repeat, but with a manual close instead of context management
        mock_broker.reset()
        s = stream.open(broker_url, "w")
        s.write_raw(encoded_msg, canonical_headers)
        s.close()
        assert mock_broker.has_message(topic, encoded_msg, canonical_headers)

        with stream.open("kafka://localhost:9092/topic1,topic2", "w") as s:
            with pytest.raises(Exception):
                s.write_raw(encoded_msg, canonical_headers)
            s.write_raw(encoded_msg, canonical_headers, topic="topic1")
            assert mock_broker.has_message("topic1", encoded_msg, canonical_headers)
            s.write_raw(encoded_msg, canonical_headers, topic="topic2")
            assert mock_broker.has_message("topic2", encoded_msg, canonical_headers)


def test_stream_auth(auth_config, tmpdir):
    # turning off authentication should give None for the auth property
    s1 = io.Stream(auth=False)
    assert s1.auth is None

    # turning on authentication should give an auth object with the data read from the default file
    with temp_config(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        s2 = io.Stream(auth=True)
        a2 = s2.auth[0]
        assert a2._config["sasl.username"] == "username"
        assert a2._config["sasl.password"] == "password"
        assert a2.username == "username"

    # turning on authentication should fail when the default file does not exist
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)), pytest.raises(FileNotFoundError):
        s3 = io.Stream(auth=True)
        a3 = s3.auth

    # anything other than a bool passed to the Stream constructor should get handed back unchanged
    s4 = io.Stream(auth="blarg")
    assert s4.auth == "blarg"


def test_stream_open(auth_config, mock_broker, mock_producer, tmpdir):
    stream = io.Stream(auth=False)

    # verify only read/writes are allowed
    with pytest.raises(ValueError) as err:
        stream.open("kafka://localhost:9092/topic1", "q")
    assert "mode must be either 'w' or 'r'" in err.value.args

    # verify that URLs with no scheme are rejected
    with pytest.raises(ValueError) as err:
        stream.open("bad://example.com/topic", "r")
    assert "invalid kafka URL: must start with 'kafka://'" in err.value.args

    # verify that URLs with no topic are rejected when reading
    with pytest.raises(ValueError) as err:
        stream.open("kafka://example.com/", "r")
    assert "no topic(s) specified in kafka URL" in err.value.args

    # verify that URLs with too many hostnames
    with pytest.raises(ValueError) as err:
        stream.open("kafka://example.com,example.net/topic", "r")
        assert "Multiple broker addresses are not supported" in err.value.args

    # verify that complete URLs are accepted
    mb = mock_broker
    with temp_config(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("hop.io.producer.Producer", side_effect=lambda c: mock_producer(mb, c.topic)), \
            patch("adc.consumer.Consumer.subscribe", MagicMock()) as subscribe:
        stream = io.Stream()
        # opening a valid URL for reading should succeed
        consumer = stream.open("kafka://example.com/topic", "r")
        # an appropriate consumer group name should be derived from the username in the auth
        assert consumer._consumer.conf.group_id.startswith(stream.auth[0].username)
        # the target topic should be subscribed to
        subscribe.assert_called_once_with(["topic"])

        # opening a valid URL for writing should succeed
        producer = stream.open("kafka://example.com/topic", "w")
        producer.write("data")


def test_stream_open_ambiguous_creds():
    creds = [
        Auth("user1", "pass1", host="example.com"),
        Auth("user2", "pass2", host="example.com"),
    ]
    stream = io.Stream(auth=creds)
    with pytest.raises(AmbiguousCredentialError) as err:
        stream.open("kafka://example.com/topic")
    assert "To select a specific credential" in err.value.args[0]


def test_unpack(circular_msg):
    kafka_msg = make_message_standard(circular_msg)

    unpacked_msg = io.Consumer._unpack(kafka_msg)
    assert unpacked_msg == circular_msg

    unpacked_msg2, metadata = io.Consumer._unpack(kafka_msg, metadata=True)
    assert unpacked_msg2 == unpacked_msg


def test_mark_done(circular_msg):
    start_at = io.StartPosition.EARLIEST
    message_data = {"format": "circular", "content": circular_msg}

    mock_message = make_message_standard(circular_msg)
    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[mock_message])
    stream = io.Stream(until_eos=True, start_at=start_at, auth=False)

    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
        with stream.open("kafka://hostname:port/gcn", "r") as s:
            for msg, metadata in s.read(metadata=True):
                s.mark_done(metadata)
                mock_instance.mark_done.assert_called()


def test_pack(circular_msg, circular_text):
    # message class
    packed_msg, headers = io.Producer.pack(circular_msg)

    # unstructured message
    message = {"hey": "you"}
    packed, headers = io.Producer.pack(message)


@pytest.mark.parametrize("message", [
    {"format": "voevent", "read-mode": "rb"},
    {"format": "circular", "read-mode": "r"},
    {"format": "blob", "read-mode": "rb"},
    {"format": "json", "read-mode": "rb"},
    {"format": "avro", "read-mode": "rb"},
])
def test_pack_unpack_roundtrip(message, message_parameters_dict, caplog):
    format = message["format"]

    # load test data
    shared_datadir = Path("tests/data")
    test_filename = message_parameters_dict[format]["test_file"]
    test_file = shared_datadir / "test_data" / test_filename

    # generate a message
    expected_model = message_parameters_dict[format]["expected_model"]
    with open(test_file, message["read-mode"]) as f:
        orig_message = expected_model.load(f)

    # pack the message
    packed_msg, headers = io.Producer.pack(orig_message)

    # mock a kafka message with value being the packed message
    kafka_msg = make_message(packed_msg, headers)

    # unpack the message
    unpacked_msg = io.Consumer._unpack(kafka_msg)

    # verify based on format
    if format in ("voevent", "circular"):
        assert isinstance(unpacked_msg, expected_model)
        assert unpacked_msg == orig_message
    else:
        assert isinstance(unpacked_msg, type(unpacked_msg))
        assert unpacked_msg == orig_message


def test_pack_unpack_roundtrip_unstructured():
    # objects (of types that json.loads happens to produce, and bytes) should remain unchanged
    # by the process of packing and unpacking
    for orig_message in [
            "a string",
            ["a", "B", "c"],
            {"dict": True, "other_data": [5, 17]},
            b"A\x00B\x04"]:
        packed_msg, headers = io.Producer.pack(orig_message)
        print(f"pasked_msg: {packed_msg}")
        kafka_msg = make_message(packed_msg, headers)
        unpacked_msg = io.Consumer._unpack(kafka_msg)
        assert unpacked_msg.content == orig_message

    # non-serializable objects should raise an error
    with pytest.raises(TypeError):
        # note that we are not trying to pack a string, but the string class itself
        packed_msg, _ = io.Producer.pack(str)


@pytest.mark.parametrize("message", [
    {"format": "voevent", "content": content_mock(VOEvent)},
    {"format": "circular", "content": content_mock(GCNCircular)}
])
def test_load_load_file_equivalence(message, message_parameters_dict):
    format = message["format"]
    content = message["content"]

    # load test data
    shared_datadir = Path("tests/data")
    test_filename = message_parameters_dict[format]["test_file"]
    test_file = shared_datadir / "test_data" / test_filename

    deserializer = io.Deserializer[format.upper()]
    from_file = deserializer.load_file(test_file)

    with open(test_file, "r") as df:
        raw_data = df.read()
    from_mem = deserializer.load(raw_data)
    assert from_mem == from_file


def test_metadata(mock_kafka_message):
    metadata = io.Metadata.from_message(mock_kafka_message)

    # verify all properties are populated and match raw message
    assert metadata._raw == mock_kafka_message
    assert metadata.topic == mock_kafka_message.topic()
    assert metadata.partition == mock_kafka_message.partition()
    assert metadata.offset == mock_kafka_message.offset()
    assert metadata.timestamp == mock_kafka_message.timestamp()[1]
    assert metadata.key == mock_kafka_message.key()
    assert metadata.headers == mock_kafka_message.headers()


def test_plugin_loading(caplog):
    # plugins which fail during loading should trigger a warning
    import pluggy
    pm1 = pluggy.PluginManager("hop")

    def raise_ex(ignored):
        raise Exception("Things are bad")
    lse_mock = MagicMock(side_effect=raise_ex)
    pm1.load_setuptools_entrypoints = lse_mock

    builtin_models = ["VOEVENT", "GCNTEXTNOTICE", "CIRCULAR", "JSON", "AVRO", "BLOB"]

    with patch("pluggy.PluginManager", MagicMock(return_value=pm1)), \
            caplog.at_level(logging.WARNING):
        registered = io._load_deserializer_plugins()
        assert "Could not load external message plugins" in caplog.text
        # but built-in models should still be available
        assert len(registered) == len(builtin_models)
        for expected_model in builtin_models:
            assert expected_model in registered
        # while we're here, make sure the documented interface is being used
        lse_mock.assert_called_with("hop_plugin")

    # users should be warned if plugins collide
    pm2 = pluggy.PluginManager("hop")
    gm_mock = MagicMock(return_value=[{
                        "foo": None,
                        "bar": None,
                        }, {"Foo": str}])
    # If we don't make this property explicitly false, the Mock will helpfully create it as another
    # Mock object which will then look true-ish to pluggy, leading to problems
    gm_mock.spec.warn_on_impl = False
    pm2.hook.get_models = gm_mock
    with patch("pluggy.PluginManager", MagicMock(return_value=pm2)), \
            caplog.at_level(logging.WARNING):
        registered = io._load_deserializer_plugins()
        assert "Identified duplicate message plugin" in caplog.text

    # additional plugins should appear in the resulting list
    pm3 = pluggy.PluginManager("hop")

    def fake_lse(ignored):
        original_get_models = pm3.hook.get_models

        def add_extra_model():
            models = original_get_models()
            models.append({"Foo": None})
            return models
        pm3.hook.get_models = add_extra_model
    pm3.load_setuptools_entrypoints = fake_lse
    with patch("pluggy.PluginManager", MagicMock(return_value=pm3)):
        registered = io._load_deserializer_plugins()
        assert len(registered) == len(builtin_models) + 1
        for expected_model in builtin_models:
            assert expected_model in registered
        assert "FOO" in registered


def test_stream_flush():
    # flush is pretty trivial; it should just call the underlying flush
    with patch("adc.producer.Producer.flush", MagicMock()) as flush:
        broker_url = "kafka://example.com:9092/topic"
        stream = io.Stream(auth=False).open(broker_url, "w")
        stream.flush()
        flush.assert_called()


def test_is_test(circular_msg):
    fake_message = MagicMock()

    def fake_headers_none():
        return None

    def fake_headers_list_test():
        return [('_test', b'true')]

    def fake_headers_list():
        return [('foo', b'bar')]

    fake_message.headers = fake_headers_none
    ret = io.Consumer.is_test(fake_message)
    assert ret is False
    fake_message.headers = fake_headers_list_test
    ret = io.Consumer.is_test(fake_message)
    assert ret is True
    fake_message.headers = fake_headers_list
    ret = io.Consumer.is_test(fake_message)
    assert ret is False


def make_mock_listing_consumer(topics=[]):
    """Create a mock object suitable for replacing confluent_kafka.Consumer
        which has a list_topics method which acts as if a predetermined set of
        topics exist.
    """
    def get_topics(topic=None, timeout=None):
        nonlocal topics
        result = {}
        if topic is None:
            for topic in topics:
                result[topic] = MagicMock()  # don't care much what value is
                result[topic].error = None  # but it should claim no errors
        elif topic in topics:
            result[topic] = MagicMock()
            result[topic].error = None
        # wrap in a fake metadata object
        metadata = MagicMock()
        metadata.topics = result
        return metadata
    consumer = MagicMock()
    consumer.list_topics = get_topics
    return MagicMock(return_value=consumer)


def test_list_topics():
    # when there are some topics, they are all listed
    with patch("confluent_kafka.Consumer", make_mock_listing_consumer(["foo", "bar"])) as Consumer:
        listing = io.list_topics("kafka://example.com", auth=False)
        assert len(listing) == 2
        assert "foo" in listing
        assert "bar" in listing

        # when auth=False, no auth related properties are set
        Consumer.assert_called_once()
        cons_args = Consumer.call_args[0]
        assert len(cons_args) == 1
        assert "sasl.username" not in cons_args[0]
        assert "sasl.password" not in cons_args[0]

    # when there are no topics, the result is empty
    with patch("confluent_kafka.Consumer", make_mock_listing_consumer([])) as Consumer:
        listing = io.list_topics("kafka://example.com", auth=False)
        assert len(listing) == 0

    # result topics should be the intersection of ones which exist and which were requested
    with patch("confluent_kafka.Consumer", make_mock_listing_consumer(["foo", "bar"])) as Consumer:
        listing = io.list_topics("kafka://example.com/bar,baz", auth=False)
        assert len(listing) == 1
        assert "bar" in listing


def test_list_topics_too_many_brokers():
    # If the URL specifies too many brokers, an error is raised
    with patch("confluent_kafka.Consumer", make_mock_listing_consumer([])):
        with pytest.raises(ValueError):
            listing = io.list_topics("kafka://example.com,example.net", auth=False)


def test_list_topics_auth(auth_config, tmpdir):
    # when auth=True, auth related properties should be set
    with temp_config(tmpdir, auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("confluent_kafka.Consumer", make_mock_listing_consumer([])) as Consumer:
        listing = io.list_topics("kafka://example.com", auth=True)

        # when auth=True, auth related properties should be set
        Consumer.assert_called_once()
        cons_args = Consumer.call_args[0]
        assert len(cons_args) == 1
        assert "sasl.username" in cons_args[0]
        assert "sasl.password" in cons_args[0]
        assert cons_args[0]["sasl.username"] == "username"
        assert cons_args[0]["sasl.password"] == "password"

    # when an Auth object is set, it should take precedence over automatic lookup
    with temp_config(tmpdir, auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("confluent_kafka.Consumer", make_mock_listing_consumer([])) as Consumer:
        auth = Auth("someone_else", "other_password")
        listing = io.list_topics("kafka://example.com", auth=auth)

        # when auth=Auth(...), auth related properties should be set correctly
        Consumer.assert_called_once()
        cons_args = Consumer.call_args[0]
        assert len(cons_args) == 1
        assert "sasl.username" in cons_args[0]
        assert "sasl.password" in cons_args[0]
        assert cons_args[0]["sasl.username"] == auth.username
        assert cons_args[0]["sasl.password"] == auth.password

    # when an Auth object is set, it should take precedence over automatic lookup,
    # even with userinfo in the URL
    with temp_config(tmpdir, auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("confluent_kafka.Consumer", make_mock_listing_consumer([])) as Consumer:
        auth = Auth("someone_else", "other_password")
        listing = io.list_topics("kafka://username@example.com", auth=auth)

        # when auth=Auth(...), auth related properties should be set correctly
        Consumer.assert_called_once()
        cons_args = Consumer.call_args[0]
        assert len(cons_args) == 1
        assert "sasl.username" in cons_args[0]
        assert "sasl.password" in cons_args[0]
        assert cons_args[0]["sasl.username"] == auth.username
        assert cons_args[0]["sasl.password"] == auth.password

    # when auth=True and userinfo, the correct credential should be automatically
    # looked up
    multi_cred = """
        auth = [{
            username="user1",
            password="pass1"
            },
            {
            username="user2",
            password="pass2"
            }]
        """
    with temp_config(tmpdir, multi_cred) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("confluent_kafka.Consumer", make_mock_listing_consumer([])) as Consumer:
        listing = io.list_topics("kafka://user2@example.com", auth=True)

        Consumer.assert_called_once()
        cons_args = Consumer.call_args[0]
        assert len(cons_args) == 1
        assert "sasl.username" in cons_args[0]
        assert "sasl.password" in cons_args[0]
        assert cons_args[0]["sasl.username"] == "user2"
        assert cons_args[0]["sasl.password"] == "pass2"


def test_list_topics_timeout():
    cred = Auth("user", "pass")
    for timeout in [0.5, 1.0, 1.5]:
        start = time.time()
        with pytest.raises(confluent_kafka.KafkaException) as err:
            io.list_topics("kafka://not-a-valid-broker.scimma.org", auth=cred, timeout=timeout)
        stop = time.time()
        assert abs((stop - start) - timeout) < 0.1



from collections.abc import Mapping
from dataclasses import dataclass
from functools import lru_cache
from enum import Enum
import json
import logging
import random
import string
from typing import List, Tuple, Union
from uuid import uuid4
import warnings

import confluent_kafka
import pluggy

from adc import consumer, errors, kafka, producer

from .configure import get_config_path
from .auth import Auth, AmbiguousCredentialError
from .auth import load_auth
from .auth import select_matching_auth
from . import models
from . import plugins

logger = logging.getLogger("hop")

StartPosition = consumer.ConsumerStartPosition


class Stream(object):
    """Defines an event stream.

    Sets up defaults used within the client so that when a
    stream connection is opened, it will use defaults specified here.

    Args:
        auth: A `bool` or :class:`Auth <hop.auth.Auth>` instance. Defaults to
            loading from :meth:`auth.load_auth <hop.auth.load_auth>` if set to
            True. To disable authentication, set to False.
        start_at: The message offset to start at in read mode. Defaults to LATEST.
        until_eos: Whether to listen to new messages forever (False) or stop
            when EOS is received in read mode (True). Defaults to False.

    """

    def __init__(self, auth=True, start_at=StartPosition.LATEST, until_eos=False):
        self._auth = [auth] if isinstance(auth, Auth) else auth
        self.start_at = start_at
        self.until_eos = until_eos

    @property
    @lru_cache(maxsize=1)
    def auth(self):
        # configuration is disabled in adc-streaming by passing None,
        # so to provide a nicer interface, we allow boolean flags as well.
        # this also explicitly gets around a problem in setting
        # configuration to True by default in the convenience class `stream`
        # which is set to `Stream()`. instead, configuration is first loaded
        # during the first open stream and cached for future use.
        if isinstance(self._auth, bool):
            if self._auth:
                try:
                    return load_auth()
                except FileNotFoundError:
                    logger.error(
                        "configuration set to True and configuration file "
                        f"not found at {get_config_path('auth')} to authenticate"
                    )
                    raise
            else:
                return None
        else:
            return self._auth

    def open(self, url, mode="r", group_id=None, ignoretest=True, **kwargs):
        """Opens a connection to an event stream.

        Args:
            url: Sets the broker URL to connect to.
            mode: Read ('r') or write ('w') from the stream.
            group_id: The consumer group ID from which to read.
                      Generated automatically if not specified.
            ignoretest: When True, read mode will silently discard
                        test messages.

        Returns:
            An open connection to the client, either a :class:`Producer` instance
            in write mode or a :class:`Consumer` instance in read mode.

        Raises:
            ValueError: If the mode is not set to read/write, if no topic
                is specified in read mode, or if more than one broker is specified

        """
        username, broker_addresses, topics = kafka.parse_kafka_url(url)
        if len(broker_addresses) > 1:
            raise ValueError("Multiple broker addresses are not supported")
        logger.debug("connecting to addresses=%s  username=%s  topics=%s",
                     broker_addresses, group_id, topics)

        if self.auth is not None:
            try:
                credential = select_matching_auth(self.auth, broker_addresses[0], username)
            except AmbiguousCredentialError as err:
                msg = err.message
                msg += "\nTo select a specific credential to use, embed its username in the URL:"
                msg += f"\n  kafka://<username>@{broker_addresses[0]}/{','.join(topics)}"
                raise AmbiguousCredentialError(msg)
        else:
            credential = None

        if mode == "w":
            if topics is None or len(topics) != 1:
                topics = [None]
            if group_id is not None:
                warnings.warn("group ID has no effect when opening a stream in write mode")
            return Producer(broker_addresses, topics[0], auth=credential, **kwargs)
        elif mode == "r":
            if topics is None or len(topics) == 0:
                raise ValueError("no topic(s) specified in kafka URL")
            if group_id is None:
                username = credential.username if credential is not None else None
                group_id = _generate_group_id(username, 10)
                logger.info(f"group ID not specified, generating a random group ID: {group_id}")
            return Consumer(
                group_id,
                broker_addresses,
                topics,
                start_at=self.start_at,
                auth=credential,
                read_forever=not self.until_eos,
                ignoretest=ignoretest,
                **kwargs,
            )
        else:
            raise ValueError("mode must be either 'w' or 'r'")


def _load_deserializer_plugins():
    """Load all registered deserializer plugins.

    """
    # set up plugin manager
    manager = pluggy.PluginManager("hop")
    manager.add_hookspecs(plugins)

    # load in base models
    manager.register(models)

    # load external models
    try:
        manager.load_setuptools_entrypoints("hop_plugin")
    except Exception:
        logger.warning(
            "Could not load external message plugins as one or more plugins "
            "generated errors upon import. to fix this issue, uninstall or fix "
            "any problematic plugins that are currently installed."
        )
        import sys
        import traceback
        traceback.print_exc(file=sys.stderr)

    # add all registered plugins to registry
    registered = {}
    for model_plugins in manager.hook.get_models():
        for name, model in model_plugins.items():
            plugin_name = name.upper()
            if plugin_name in registered:
                logger.warning(
                    f"Identified duplicate message plugin {plugin_name} registered under "
                    "the same name. this may cause unexpected behavior when using this "
                    "message format."
                )
            registered[plugin_name] = model

    return registered


class _DeserializerMixin:
    @classmethod
    def deserialize(cls, message):
        """Deserialize a stream message and instantiate a model.

        Args:
            message: A serialized message.

        Returns:
            A data container corresponding to the format in the
            serialized message.

        Raises:
            ValueError: If the message is incorrectly formatted or
                if the message format is not recognized.

        """

        def from_format(data, format, deserialize=True):
            format = format.upper()
            if format in cls.__members__:
                if deserialize:
                    return cls[format].value.deserialize(data)
                else:
                    return cls[format].value(**data)
            else:
                logger.warning(f"Message format {format} not recognized; returning a Blob")
                logger.warning(f"Known message formats are {cls.__members__}")
                return models.Blob(content=data)

        # first look for a format header
        if message.headers() is not None:
            try:
                format_header = filter(lambda h: h[0] == "_format", message.headers())
                format = next(format_header)[1].decode("utf-8")
                return from_format(message.value(), format)
            except StopIteration:  # no format header
                pass
        # otherwise, try doing old-style JSON envelope decoding
        try:
            old = json.loads(message.value().decode("utf-8"))
            if isinstance(old, Mapping) and "format" in old and "content" in old:
                if old["format"] == "blob":  # this was the old label for JSON
                    return models.JSONBlob(content=old["content"])
                return from_format(old["content"], old["format"], deserialize=False)
            # not labeled according to our scheme, but it is valid JSON
            return models.JSONBlob(content=old)
        # if we can't tell what the data is, pass it on unchanged
        except (UnicodeDecodeError, json.JSONDecodeError):
            logger.info("Unknown message format; returning a Blob")
            return models.Blob(content=message.value())

    def load(self, input_):
        return self.value.load(input_)

    def load_file(self, input_file):
        return self.value.load_file(input_file)


Deserializer = Enum(
    "Deserializer",
    _load_deserializer_plugins(),
    module=__name__,
    type=_DeserializerMixin
)


def _generate_group_id(user, n):
    """Generate a random Kafka group ID.

    Args:
        user: Username associated with the credential being used
        n: Length of randomly generated string suffix.

    Returns:
        The generated group ID.

    """
    alphanum = string.ascii_uppercase + string.digits
    rand_str = ''.join(random.SystemRandom().choice(alphanum) for _ in range(n))
    if user is None:
        return rand_str
    return '-'.join((user, rand_str))


@dataclass(frozen=True)
class Metadata:
    """Broker-specific metadata that accompanies a consumed message.

    """

    topic: str
    partition: int
    offset: int
    timestamp: int
    key: Union[str, bytes]
    headers: List[Tuple[str, bytes]]
    _raw: confluent_kafka.Message

    @classmethod
    def from_message(cls, msg: confluent_kafka.Message) -> 'Metadata':
        return cls(
            topic=msg.topic(),
            partition=msg.partition(),
            offset=msg.offset(),
            timestamp=msg.timestamp()[1],
            key=msg.key(),
            headers=msg.headers(),
            _raw=msg,
        )


class Consumer:
    """
    An event stream opened for reading one or more topics.
    Instances of this class should be obtained from :meth:`Stream.open`.
    """

    def __init__(self, group_id, broker_addresses, topics, ignoretest=True, **kwargs):
        """
        Args:
            group_id: The Kafka consumer group to join for reading messages.
            broker_addresses: The list of bootstrap Kafka broker URLs.
            topics: The list of names of topics to which to subscribe.
            read_forever: If true, keep the stream open to wait for more messages
                after reading the last currently available message.
            start_at: The position in the topic stream at which to start
                reading, specified as a StartPosition object.
            auth: An adc.auth.SASLAuth object specifying client authentication
                to use.
            error_callback: A callback which will be called with any
                confluent_kafka.KafkaError objects produced representing internal
                Kafka errors.
            offset_commit_interval: A datetime.timedelta specifying how often to
                report progress to Kafka.
            ignoretest: When True, ignore test messages. When False, process them
                normally.

        :meta private:
        """
        logger.info(f"connecting to kafka://{','.join(broker_addresses)}")
        self._consumer = consumer.Consumer(consumer.ConsumerConfig(
            broker_urls=broker_addresses,
            group_id=group_id,
            **kwargs,
        ))
        logger.info(f"subscribing to topics: {topics}")
        self._consumer.subscribe(topics)
        self.ignoretest = ignoretest

    def read(self, metadata=False, autocommit=True, **kwargs):
        """Read messages from a stream.

        Args:
            metadata: Whether to receive message metadata alongside messages.
            autocommit: Whether messages are automatically marked as handled
                via `mark_done` when the next message is yielded. Defaults to True.
            batch_size: The number of messages to request from Kafka at a time.
                Lower numbers can give lower latency, while higher numbers will
                be more efficient, but may add latency.
            batch_timeout: The period of time to wait to get a full batch of
                messages from Kafka. Similar to batch_size, lower numbers can
                reduce latency while higher numbers can be more efficient at the
                cost of greater latency.
                If specified, this argument should be a datetime.timedelta
                object.
        """
        logger.info("processing messages from stream")
        for message in self._consumer.stream(autocommit=autocommit, **kwargs):
            if self.ignoretest and self.is_test(message):
                continue
            yield self._unpack(message, metadata=metadata)
        logger.info("finished processing messages")

    @staticmethod
    def _unpack(message, metadata=False):
        """Deserialize and unpack messages.

        Args:
            message: The message to deserialize and unpack.
            metadata: Whether to receive message metadata alongside messages.
        """
        payload = Deserializer.deserialize(message)
        if metadata:
            return (payload, Metadata.from_message(message))
        else:
            return payload

    def read_raw(self, metadata=False, autocommit=True, **kwargs):
        """Read messages from a stream without applying any deserialization.

        This is an advanced interface; for most purposes it is preferrable to use
        :meth:`Consumer.read <hop.io.Consumer.read>` instead.

        Args:
            metadata: Whether to receive message metadata alongside messages.
            autocommit: Whether messages are automatically marked as handled
                via `mark_done` when the next message is yielded. Defaults to True.
            batch_size: The number of messages to request from Kafka at a time.
                Lower numbers can give lower latency, while higher numbers will
                be more efficient, but may add latency.
            batch_timeout: The period of time to wait to get a full batch of
                messages from Kafka. Similar to batch_size, lower numbers can
                reduce latency while higher numbers can be more efficient at the
                cost of greater latency.
                If specified, this argument should be a datetime.timedelta
                object.
        """
        logger.info("processing messages from stream")
        for message in self._consumer.stream(autocommit=autocommit, **kwargs):
            if self.ignoretest and self.is_test(message):
                continue
            payload = message.value()
            if metadata:
                yield (payload, Metadata.from_message(message))
            else:
                yield payload
        logger.info("finished processing messages")

    def mark_done(self, metadata, asynchronous: bool = True):
        """Mark a message as fully-processed.

        Args:
            metadata: A Metadata instance containing broker-specific metadata.
            asynchronous: Whether to allow the commit to happen asynchronously
                          in the background.

        """
        self._consumer.mark_done(metadata._raw, asynchronous)

    def stop(self):
        """Stops the runloop of the consumer. Useful when running the
        consumer in a different thread.
        """
        self._consumer.stop()

    def close(self):
        """End all subscriptions and shut down.

        """
        logger.info("closing connection")
        self._consumer.close()

    @staticmethod
    def is_test(message):
        """True if message is a test message (contains '_test' as a header key).

        Args:
            message: The message to test.
        """
        h = message.headers()
        if h is None:
            return False
        else:
            return bool([v for k, v in h if k == "_test"])

    def __iter__(self):
        yield from self.read()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()


class Producer:
    """
    An event stream opened for writing to a topic.
    Instances of this class should be obtained from :meth:`Stream.open`.
    """

    def __init__(self, broker_addresses, topic, auth, **kwargs):
        """
        Args:
            broker_addresses: The list of bootstrap Kafka broker URLs.
            topic: The name of the topic to which to write.
            auth: An adc.auth.SASLAuth object specifying client authentication
                to use.
            error_callback: A callback which will be called with any
                confluent_kafka.KafkaError objects produced representing internal
                Kafka errors.
            produce_timeout: A datetime.timedelta object specifying the maximum
                time to wait for a message to be sent to Kafka. If zero, sending
                will never time out.

        :meta private:
        """
        logger.info(f"connecting to kafka://{','.join(broker_addresses)}")
        self._producer = producer.Producer(producer.ProducerConfig(
            broker_urls=broker_addresses,
            topic=topic,
            auth=auth,
            **kwargs,
        ))
        self.auth = auth
        logger.info(f"publishing to topic: {topic}")

    def write(self, message, headers=None,
              delivery_callback=errors.raise_delivery_errors, test=False, topic=None):
        """Write messages to a stream.


        Args:
            message: The message to write.
            headers: The set of headers requested to be sent with the message, either as a
                     mapping, or as a list of 2-tuples. In either the mapping or the list case,
                     all header keys must be strings and and values should be either string-like or
                     bytes-like objects.
            delivery_callback: A callback which will be called when each message
                is either delivered or permenantly fails to be delivered.
            test: Message should be marked as a test message by adding a header
                  with key '_test'.
            topic: The topic to which the message should be sent. This need not be specified if
                   the stream was opened with a URL containing exactly one topic name.
        """
        message, headers = self._pack(message, headers, test=test)
        self._producer.write(message, headers=headers, delivery_callback=delivery_callback,
                             topic=topic)

    def write_raw(self, packed_message, headers=None,
                  delivery_callback=errors.raise_delivery_errors, topic=None):
        """Write a pre-encoded message to the stream.

        This is an advanced interface; for most purposes it is preferrable to use
        :meth:`Producer.write <hop.io.Producer.write>` instead.

        Args:
            packed_message: The message to write, which must already be correctly encoded by
                            :meth:`Producer.pack <hop.io.Producer.pack>`
            headers: Any headers to attach to the message, either as a dictionary
                mapping strings to strings, or as a list of 2-tuples of strings.
            delivery_callback: A callback which will be called when each message
                is either delivered or permenantly fails to be delivered.
            topic: The topic to which the message should be sent. This need not be specified if
                   the stream was opened with a URL containing exactly one topic name.
        """

        self._producer.write(packed_message, headers=headers, delivery_callback=delivery_callback,
                             topic=topic)

    @staticmethod
    def pack(message, headers=None, test=False, auth=None):
        """Pack and serialize a message.

        This is an advanced interface, which most users should not need to call directly, as
        :meth:`Producer.write <hop.io.Producer.write>` uses it automatically.

        Args:
            message: The message to pack and serialize.
            headers: The set of headers requested to be sent with the message, either as a
                     mapping, or as a list of 2-tuples. In either the mapping or the list case,
                     all header keys must be strings and and values should be either string-like or
                     bytes-like objects.
            test: Message should be marked as a test message by adding a header
                  with key '_test'.

        Returns:
            A tuple containing the serialized message and the collection of headers which
            should be sent with it.

        """
        # canonicalize headers to list form
        if headers is None:
            headers = []
        elif isinstance(headers, Mapping):
            headers = list(headers.items())
        # Assign a UUID to the message
        headers.append(("_id", uuid4().bytes))
        if auth is not None:
            headers.append(("_sender", auth.username.encode("utf-8")))
        if test:
            headers.append(('_test', b"true"))
        try:  # first try telling the message to serialize itself
            encoded = message.serialize()
            headers.append(("_format", encoded["format"].encode("utf-8")))
            payload = encoded["content"]
        except AttributeError:  # message is not a model object which can self-serialize
            try:  # serialize the message as JSON if possible
                blob = models.JSONBlob(content=message)
                encoded = blob.serialize()
                headers.append(("_format", encoded["format"].encode("utf-8")))
                payload = encoded["content"]
            except TypeError:  # message can't be turned into JSON
                if isinstance(message, bytes):
                    blob = models.Blob(content=message)
                    encoded = blob.serialize()
                    headers.append(("_format", encoded["format"].encode("utf-8")))
                    payload = encoded["content"]
                else:
                    raise TypeError("Unable to pack a message of type "
                                    + message.__class__.__name__
                                    + " which is not bytes and cannot be serialized to JSON")
        return (payload, headers)

    def _pack(self, message, headers, test):
        """Internal wrapper for :meth:`pack <hop.io.Producer.pack>` which
        automatically sets the auth parameter.
        """
        return self.pack(message, headers=headers, test=test, auth=self.auth)

    def flush(self):
        """Request that any messages locally queued for sending be sent immediately.
        """
        self._producer.flush()

    def close(self):
        """Wait for enqueued messages to be written and shut down.

        """
        logger.info("closing connection")
        return self._producer.close()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()
        return self._producer.__exit__(*exc)


def list_topics(url: str, auth: Union[bool, Auth] = True, timeout=-1.0):
    """List the accessible topics on the Kafka broker referred to by url.

    Args:
        url: The Kafka broker URL. Only one broker may be specified. Topics
            may be specified, in which case only topics in the intersection
            of the set specified by the URL and actually present on the broker
            will be returned. If a userinfo component is present in the URL and
            auth is True, it will be treated as a hint to the automatic auth
            lookup.
        auth: A `bool` or :class:`Auth <hop.auth.Auth>` instance. Defaults to
            loading from :meth:`auth.load_auth <hop.auth.load_auth>` if set to
            True. To disable authentication, set to False. If a username is
            specified as part of url but auth is a :class:`Auth <hop.auth.Auth>`
            instance the url information will be ignored.
        timeout: A floating point value, indicating the maximum number of
            seconds to wait to connect to a broker, or a negative value to
            indicate no limit.

    Returns:
        A dictionary mapping topic names to
        :class:`confluent_kafka.admin.TopicMetadata` instances.

    Raises:
        ValueError: If more than one broker is specified.
        confluent_kafka.KafkaException: If connecting to the broker times out.
    """
    username, broker_addresses, query_topics = kafka.parse_kafka_url(url)
    if len(broker_addresses) > 1:
        raise ValueError("Multiple broker addresses are not supported")
    user_auth = None
    if auth is True:
        credentials = load_auth()
        user_auth = select_matching_auth(credentials, broker_addresses[0], username)
    elif auth is not False:
        user_auth = auth
    group_id = _generate_group_id(username, 10)
    config = {
        "bootstrap.servers": ",".join(broker_addresses),
        "error_cb": errors.log_client_errors,
        "group.id": group_id,
    }
    if user_auth is not None:
        config.update(user_auth())
    consumer = confluent_kafka.Consumer(config)
    valid_topics = {}
    if query_topics is not None:
        for topic in query_topics:
            topic_data = consumer.list_topics(topic=topic, timeout=timeout).topics
            for topic in topic_data.keys():
                if topic_data[topic].error is None:
                    valid_topics[topic] = topic_data[topic]
    else:
        topic_data = consumer.list_topics(timeout=timeout).topics
        valid_topics = {t: d for t, d in topic_data.items() if d.error is None}
    return valid_topics



from unittest.mock import patch, mock_open
import json
import pytest

from hop import models


def test_MessageModel():
    # derived classes must implement their own load() logic, and the base implementation has no
    # useful functionality
    with pytest.raises(NotImplementedError):
        models.MessageModel.load(None)


def test_voevent(voevent_fileobj):
    voevent = models.VOEvent.load(voevent_fileobj)

    # check a few attributes
    assert voevent.ivorn == "ivo://gwnet/LVC#S200302c-1-Preliminary"
    assert voevent.role == "observation"
    assert voevent.version == "2.0"

    assert voevent.Who["Date"] == "2020-03-02T02:00:09"
    assert voevent.Description == "Report of a candidate gravitational wave event"
    assert voevent.WhereWhen["ObsDataLocation"]["ObservatoryLocation"]["id"] == "LIGO Virgo"

    # verify wrapper format is correct
    assert voevent.serialize()["format"] == "voevent"

    # conversion to a string should produce valid JSON which should parse to match the original
    # event data
    as_string = str(voevent)
    reparsed = json.loads(as_string)
    assert reparsed["ivorn"] == voevent.ivorn
    assert reparsed["role"] == voevent.role
    assert reparsed["version"] == voevent.version
    assert reparsed["Who"]["Date"] == voevent.Who["Date"]
    assert reparsed["Description"] == voevent.Description
    assert reparsed["WhereWhen"]["ObsDataLocation"]["ObservatoryLocation"]["id"] \
        == voevent.WhereWhen["ObsDataLocation"]["ObservatoryLocation"]["id"]


def test_gcn_text_notice(gcn_text_notice_fileobj, gcn_text_notice_data):
    text_notice = models.GCNTextNotice.load(gcn_text_notice_fileobj)

    # the object should retain the raw data unchanged for lossless round-tripping
    assert text_notice.raw == gcn_text_notice_data

    # check the data was parsed sensibly
    expected_fields = ["title", "notice_date", "notice_type", "stream", "run_num", "event_num",
                       "src_ra", "src_dec", "src_error", "src_error50", "discovery_date",
                       "discovery_time", "revision", "energy", "signalness", "far", "sun_postn",
                       "sun_dist", "moon_postn", "moon_dist", "gal_coords", "ecl_coords",
                       "comments"]
    for field in expected_fields:
        assert field in text_notice.fields

    expected_title = "GCN/AMON NOTICE"
    assert text_notice.fields["title"] == expected_title
    expected_energy = "1.2126e+02 [TeV]"
    assert text_notice.fields["energy"] == expected_energy
    # multi-line values should be accumulated
    expected_src_ra = \
        "103.7861d {+06h 55m 09s} (J2000),\n" \
        "104.1106d {+06h 56m 27s} (current),\n" \
        "103.1176d {+06h 52m 28s} (1950)"
    assert text_notice.fields["src_ra"] == expected_src_ra
    # repeated keys should have their values accumulated
    expected_comments = \
        "IceCube Bronze event.\n" \
        "The position error is statistical only, there is no systematic added."
    assert text_notice.fields["comments"] == expected_comments


def test_gcn_circular(circular_text, circular_msg, circular_data_raw):
    with patch("builtins.open", mock_open(read_data=circular_text)):
        gcn_file = "example.gcn3"
        with open(gcn_file, "r") as f:
            gcn = models.GCNCircular.load(f)

        # verify parsed GCN structure is correct
        assert gcn.header["title"] == circular_data_raw["header"]["title"]
        assert gcn.header["number"] == circular_data_raw["header"]["number"]
        assert gcn.header["subject"] == circular_data_raw["header"]["subject"]
        assert gcn.header["date"] == circular_data_raw["header"]["date"]
        assert gcn.header["from"] == circular_data_raw["header"]["from"]

        assert gcn.body == circular_data_raw["body"]

        # verify wrapper format is correct
        assert gcn.serialize()["format"] == "circular"

        # conversion to a string should produce email formatted data which should parse to match the
        # original event data
        as_string = str(gcn)
        reparsed = models.GCNCircular.load(as_string)
        assert reparsed.header == gcn.header
        assert reparsed.body == gcn.body


def test_blob(blob_text, blob_msg):
    with patch("builtins.open", mock_open(read_data=blob_text)):
        blob_file = "example_blob.txt"
        with open(blob_file, "r") as f:
            blob = models.Blob.load(f)

        # verify blob text is correct
        assert blob.content == blob_msg["content"]

        # verify wrapper format is correct
        assert blob.serialize()["format"] == "blob"

        # if the blob was holding a string, conversion to a string should just return that
        as_string = str(blob)
        assert as_string == blob.content
        # round-tripping should work iff the blob was holding a string all along
        assert models.Blob.load(as_string) == blob


def test_avro(avro_data_raw, avro_data, avro_msg):
    with patch("builtins.open", mock_open(read_data=avro_data)):
        avro = models.AvroBlob.load(avro_data)

    # verify data is correct after deserializing
    assert avro.content == avro_data_raw

    # verify wrapper format is correct
    assert avro.serialize()["format"] == "avro"

    # verify that serializing again produces the same result
    serialized = avro.serialize()["content"]
    assert models.AvroBlob.load(serialized) == avro

    # test that serializing the same object tree without a schema produces equivalent results
    avro2 = models.AvroBlob(avro_data_raw)
    serialized2 = avro2.serialize()["content"]
    # serialized2 may not be the same as avro_data because the schemas can differ
    assert models.AvroBlob.load(serialized2) == avro


def test_avro_invalid_data_type_serialization():
    # data must be a sequence of records
    with pytest.raises(TypeError) as te:
        models.AvroBlob({1: "abc", False: "def"}).serialize()
    assert "AvroBlob requires content to be a sequence of records" in str(te.value)

    # dictionary keys can only be strings
    with pytest.raises(ValueError) as ve:
        models.AvroBlob([{1: "abc", False: "def"}]).serialize()
    assert "Dictionaries with non-string keys cannot be represented as Avro" in str(ve.value)

    # arbitrary, user-defined types are not supported in Avro
    with pytest.raises(ValueError) as ve:
        models.AvroBlob([models.Blob(b"somedata")]).serialize()
    assert "Unable to assign an Avro type to value of type <class 'hop.models.Blob'>" \
        in str(ve.value)


def test_avro_invalid_data_type_deserialization():
    # only bytes objects are accepted, not strings
    with pytest.raises(TypeError):
        models.AvroBlob.load("some string")


def test_avro_operators(avro_data, avro_data_raw):
    avro_with_schema = models.AvroBlob.load(avro_data)
    avro_no_schema = models.AvroBlob(avro_data_raw)

    # equality should not consider schemas
    assert avro_with_schema == avro_no_schema
    # not equal to objects of any other type
    assert avro_with_schema != avro_data_raw
    # hashing is not implemented because python fails to implement it for dict
    with pytest.raises(NotImplementedError):
        hash(avro_with_schema)
